#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include "fmclient.h"
#include "forms.h"

extern FILE	*_fl_ps;
extern fmfonthandle use_font(int numb, float size);
extern int find(char name[]);

typedef struct {
  char name[15];        /* its name */
  void (*drawit)(int);  /* the drawing routine */
  /* BH added this 11/7/94 */
  void (*renderit)(int, int);  /* the postscript rendering routine */
  int scalable;		/* whether symbol can be scaled */
  int empty;            /* whether entry is used already */
} SYMBOL;

#define MAXSYMBOL       211
   /* Maximal number of symbols in table. Only half of them are
      used. Should be prime. */

extern SYMBOL symbols[MAXSYMBOL];      /* The symbols */

/*********
   Clipping.
*********/

void ps_set_clipping(float x,float y,float w,float h)
/* Sets the postscript clipping region */
{
  /* always to be used in combination with ps_unset_clipping(); implementation
	 uses gsave/grestore to push/pop rendering context */
  fprintf(_fl_ps, "gsave\n");
  fprintf(_fl_ps, "initclip\n");
  fprintf(_fl_ps, "newpath\n");
  fprintf(_fl_ps, "%f %f moveto\n", x, y);
  fprintf(_fl_ps, "%f %f lineto\n", x + w - 1.0, y);
  fprintf(_fl_ps, "%f %f lineto\n", x + w - 1.0, y + h - 1.0);
  fprintf(_fl_ps, "%f %f lineto\n", x, y + h - 1.0);
  fprintf(_fl_ps, "closepath clip\n");
}

void ps_unset_clipping()
/* Resets the postscript clipping region */
{
  /* always to be used in combination with ps_set_clipping(); implementation
	 uses gsave/grestore to push/pop rendering context */
  fprintf(_fl_ps, "grestore\n");
}

/*********
   Font related stuff
*********/

typedef struct {
  int		defined;	/* Whether the font is defined */
  char 		name[64];	/* Name of the font */
  int 		fnumb;		/* Number of sizes in use */
  float 	size[32];	/* different sizes in use */
  fmfonthandle 	fhandle[32];	/* handles for different sizes */
} FL_FONT;

extern FL_FONT fl_font[FL_MAXFONT];	/* The fonts used */
extern char Psfontsinuse[];
extern int	 Psfiulen;

static void ps_use_font(int numb, float size)
/* Sets current postscript font in required size */
{
	/* current implementation relies on fm font management for calculation
	   of string lengths, etc. and assumes postscript string has same length 
	   as fm string. It would be better to use the stringwidth operator at
	   run time */

	/* scaling with 1.33 is actually 96/72, or the dots-per-inch ratio
	   of screen against postscript. A hack? */
	fprintf(_fl_ps, "/%s findfont %f scalefont setfont\n", 
		fl_font[numb].name, 1.33*size);

	/* Fontnames are collected on the fly for inclusion in postscript
	   output. The (atend) construct is used. */
	if (strstr(Psfontsinuse, fl_font[numb].name) == NULL) {
		if (strlen(Psfontsinuse) == Psfiulen-1) return;
		strcat(Psfontsinuse, " ");
		strcat(Psfontsinuse, fl_font[numb].name);
	}
}

/*********
   Color Stuff.
*********/

/* refer to current forms color tables for actual color definitions */
extern short fl_red[4096], fl_green[4096], fl_blue[4096];

void ps_color(int col)
/* Sets the current postscript rendering color */
{
	/* hack: make sure 'WHITE' is white */
	if (col == 7)
		fprintf(_fl_ps, "1.0 1.0 1.0 setrgbcolor\n");
	else
		fprintf(_fl_ps, "%f %f %f setrgbcolor\n",
			fl_red[col]/255.0,fl_green[col]/255.0,fl_blue[col]/255.0);
}

/*********
   Drawing boxes.
*********/

static void vv(float x,float y)
  { fprintf(_fl_ps, "%f %f lineto\n", x, y); }

static void mv(float x,float y)
  { fprintf(_fl_ps, "%f %f moveto\n", x, y); }

void ps_line(float x, float y, float w, float h, int col)
/* Renders a line in the given color. */
{
  /* Assuming a one-unit wide linewidth, and line caps that extend half
	 the line's width beyond begin and endpoint. Linecap=2. The resulting
	 line, when projected on the x-axis, is exactly w units long, which is
	 identical to the forms' rendering in gl. */
  ps_color(col);
  fprintf(_fl_ps, "newpath\n"); mv(x+0.5,y+0.5); vv(x+w-0.5,y+h-0.5); 
  fprintf(_fl_ps, "closepath\nstroke\n");
}

void ps_rect(float x, float y, float w, float h, int col)
/* Renders a rectangle in the given color. */
{
  /* area-filling algorithm of Postscript is same as modern gl: a rectangle
	 defined w units wide is actually w units wide. Note that gl's pixel
	 grid coordinates are really pixel indices, whereas postscript's coordi-
	 nates are mathematical entities. Postscript one-unit wide square fills
	 area between grid coordinates where gl's pixel occupies a coordinate. */
  ps_color(col);
  fprintf(_fl_ps, "newpath\n");
  mv(x,y); vv(x+w,y); vv(x+w,y+h); vv(x,y+h);
  fprintf(_fl_ps, "closepath\nfill\n");
}

void ps_bound(float x, float y, float w, float h, int col)
/* Renders a boundary in the given color. */
{
  /* See notes for ps_rect and ps_line on coordinate calculations. This
	 routine fails when linewidth is not 1.0, which is generally the case
	 after a scale operator. Note that in drawing routines that do their
	 own scaling, (symbols!), a new linewidth is calculated  immediately
	 following the scale operator. gsave/grestore take care of resetting
	 such linewidths afterwards. */
  ps_color(col);
  fprintf(_fl_ps, "newpath\n");
  mv(x+0.5,y+0.5); vv(x+w-0.5,y+0.5); vv(x+w-0.5,y+h-0.5); vv(x+0.5,y+h-0.5);
  fprintf(_fl_ps, "closepath\nstroke\n");
}

void ps_rectbound(float x, float y, float w, float h, int col)
/* Renders a rectangle in the given color with a black boundary. */
{
  /* See notes for ps_rect and ps_line on coordinate calculations. This
	 routine fails when linewidth is not 1.0, which is generally the case
	 after a scale operator. */
  ps_color(col);
  fprintf(_fl_ps, "newpath\n");
  mv(x+1.0,y+1.0); vv(x+w-1.0,y+1.0); vv(x+w-1.0,y+h-1.0); vv(x+1.0,y+h-1.0);
  fprintf(_fl_ps, "closepath\nfill\n");
  ps_color(0);
  fprintf(_fl_ps, "newpath\n");
  mv(x+0.5,y+0.5); vv(x+w-0.5,y+0.5); vv(x+w-0.5,y+h-0.5); vv(x+0.5,y+h-0.5);
  fprintf(_fl_ps, "closepath\nstroke\n");
}

#define RS 15.0
#define min(a,b) ((a<b)?a:b)

void ps_rbox(int fill, float x,float y,float w,float h)
/* Renders a rounded box */
{
  float radius = min(RS, 0.4*min(w,h));

  /* Ah, this is where postscript shines. Discretization is performed
	 in the postscript rip, which results in real nice rounded boxes. */
  fprintf(_fl_ps, "newpath\n");
  fprintf(_fl_ps, "%f %f moveto\n", x+(w/2.0), y);
  fprintf(_fl_ps, "%f %f %f %f %f arcto\n", x+w, y, x+w, y+h, radius);
  fprintf(_fl_ps, "pop pop pop pop\n");
  fprintf(_fl_ps, "%f %f %f %f %f arcto\n", x+w, y+h, x, y+h, radius);
  fprintf(_fl_ps, "pop pop pop pop\n");
  fprintf(_fl_ps, "%f %f %f %f %f arcto\n", x, y+h, x, y, radius);
  fprintf(_fl_ps, "pop pop pop pop\n");
  fprintf(_fl_ps, "%f %f %f %f %f arcto\n", x, y, x+w, y, radius);
  fprintf(_fl_ps, "pop pop pop pop\n");
  fprintf(_fl_ps, "%f %f lineto\n", x+(w/2.0), y);

  if (fill) {
	  fprintf(_fl_ps, "closepath\nfill\n");
  } else {
	  fprintf(_fl_ps, "closepath\nstroke\n");
  }
}

void fl_ps_box(int renderstyle,
	int style,float x,float y,float w,float h,int c,float bw)
/* Renders a box. */
{
  /* User-visible rendering routine, which matches fl_drw_box().
	 the first argument is new and currently can be either FL_POSTSCRIPT or
	 FL_PSOUTLINE. FL_POSTSCRIPT does a best effort reproduction of what is
	 actually on the screen, including colors and filled areas. FL_PSOUTLINE
	 restricts to line-drawing renderings, which in many cases are superior
	 to POSTSCRIPT when photocopies are made, when faxes are being sent,
	 etcetera. Note that this is not very strict; that is bevelled boxes
	 actually have shaded polygons on the sides to provide the famous 3-D
	 look, but the goal is to render something that reproduces well */

  /* This routine is lengthy, but then again, it is used for almost all
	 generic forms rendering. So better do this well... */

  if (style == FL_NO_BOX) return;
  switch (renderstyle) {
  case FL_POSTSCRIPT:
	  switch (style) {
		case FL_ROUNDED_BOX:
		  ps_color(c); ps_rbox(1,x,y,w,h);
		  ps_color(0); ps_rbox(0,x,y,w,h);
		  break;
		case FL_RFLAT_BOX:
		  ps_color(c); ps_rbox(1,x,y,w,h);
		  break;
		case FL_RSHADOW_BOX:
		  /* draw the shadow */
		  ps_color(FL_RIGHT_BOUND_COL);
		  ps_set_clipping(x+bw,y-bw,w,bw+2.0);
		  ps_rbox(1,x+bw,y-bw,w,h);
		  ps_unset_clipping();
		  ps_set_clipping(x+w-1.0,y,bw+1.0,h-bw);
		  ps_rbox(1,x+bw,y-bw,w,h);
		  ps_unset_clipping();
		  ps_set_clipping(x+w+bw-RS,y,RS,RS);
		  ps_rbox(1,x+bw,y-bw,w,h);
		  ps_unset_clipping();
		  /* draw the box */
		  ps_color(c); ps_rbox(1,x,y,w,h);
		  ps_color(0); ps_rbox(0,x,y,w,h);
		  break;
		case FL_FRAME_BOX:
		  ps_color(FL_RIGHT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
		  mv(x+0.5,y+0.5); vv(x+0.5,y+h-0.5); vv(x+w-0.5,y+h-0.5); 
		  mv(x+2.5,y+1.5);vv(x+w-1.5,y+1.5);vv(x+w-1.5,y+h-2.5); 
		  fprintf(_fl_ps, "stroke\n");
		  ps_rect(x+2.,y+2.,w-4.,h-4.,c);
		  ps_bound(x+1.,y,w-1.,h-1.,FL_LEFT_BOUND_COL);
		  break;
		case FL_BORDER_BOX:
		  ps_rectbound(x,y,w,h,c); break;
		case FL_FLAT_BOX:
		  ps_rect(x,y,w,h,c); break;
		case FL_SHADOW_BOX:
		  ps_rect(x+bw,y-bw,w,bw,FL_RIGHT_BOUND_COL);
		  ps_rect(x+w,y-bw,bw,h,FL_RIGHT_BOUND_COL);
		  ps_rectbound(x,y,w,h,c);
		  break;
		case FL_UP_BOX:
		  ps_rect(x+bw+1.,y+bw+1.,w-2*bw-2.,h-2*bw-2.,c);
		  ps_rect(x+1.,y+h-bw-1.,w-2.,bw,FL_TOP_BOUND_COL);
		  ps_rect(x+1.,y+1.,w-2.,bw,FL_BOT_BOUND_COL);
		  ps_color(FL_RIGHT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
			mv(x+w-bw-1.,y+bw+1.); vv(x+w-bw-1.,y+h-bw-1.);
			vv(x+w-1.,y+h-1.); vv(x+w-1.,y+1.);
		  fprintf(_fl_ps, "closepath\nfill\n");
		  ps_color(FL_LEFT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
			mv(x+1.,y+1.); vv(x+1.,y+h-1.);
			vv(x+bw+1.,y+h-bw-1.); vv(x+bw+1.,y+bw+1.);
		  fprintf(_fl_ps, "closepath\nfill\n");
		  ps_bound(x,y,w,h,0);
		  break;
		case FL_DOWN_BOX:
		  ps_rect(x,y+h-bw,w,bw,FL_BOT_BOUND_COL);
		  ps_rect(x,y,w,bw,FL_TOP_BOUND_COL);
		  ps_color(FL_LEFT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
		  mv(x+w-bw,y+bw); vv(x+w-bw,y+h-bw); vv(x+w,y+h); vv(x+w,y);
		  fprintf(_fl_ps, "closepath\nfill\n");
		  ps_color(FL_RIGHT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
			mv(x,y); vv(x,y+h); vv(x+bw,y+h-bw); vv(x+bw,y+bw);
		  fprintf(_fl_ps, "closepath\nfill\n");
		  ps_rectbound(x+bw,y+bw,w-2*bw,h-2*bw,c);
		  break;
		default:
		  fl_error("fl_ps_box","Unknown box type");
		  break;
	  }
	  break;
  case FL_PSOUTLINE:
	  switch (style)
	  {
		case FL_ROUNDED_BOX:
		  ps_color(7); ps_rbox(1,x,y,w,h);
		  ps_color(0); ps_rbox(0,x,y,w,h);
		  break;
		case FL_RFLAT_BOX:
		  ps_color(7); ps_rbox(1,x,y,w,h);
		  break;
		case FL_RSHADOW_BOX:
		  /* draw the shadow */
		  ps_color(0);
		  ps_set_clipping(x+bw,y-bw,w,bw+2.0);
		  ps_rbox(1,x+bw,y-bw,w,h);
		  ps_unset_clipping();
		  ps_set_clipping(x+w-1.0,y,bw+1.0,h-bw);
		  ps_rbox(1,x+bw,y-bw,w,h);
		  ps_unset_clipping();
		  ps_set_clipping(x+w+bw-RS,y,RS,RS);
		  ps_rbox(1,x+bw,y-bw,w,h);
		  ps_unset_clipping();
		  /* draw the box */
		  ps_color(7); ps_rbox(1,x,y,w,h);
		  ps_color(0); ps_rbox(0,x,y,w,h);
		  break;
		case FL_FRAME_BOX:
		  ps_rect(x,y,w,h,7);
		  ps_color(0);
		  fprintf(_fl_ps, "newpath\n");
		  mv(x+0.5,y+0.5); vv(x+0.5,y+h-0.5); vv(x+w-0.5,y+h-0.5); 
		  mv(x+2.5,y+1.5);vv(x+w-1.5,y+1.5);vv(x+w-1.5,y+h-2.5); 
		  fprintf(_fl_ps, "stroke\n");
		  ps_bound(x+1.,y,w-1.,h-1.,FL_LEFT_BOUND_COL);
		  break;
		case FL_BORDER_BOX:
		  ps_rectbound(x,y,w,h,7); break;
		case FL_FLAT_BOX:
		  ps_rect(x,y,w,h,7); break;
		case FL_SHADOW_BOX:
		  ps_rect(x+bw,y-bw,w,bw,0);
		  ps_rect(x+w,y-bw,bw,h,0);
		  ps_rectbound(x,y,w,h,7);
		  break;
		case FL_UP_BOX:
		  ps_rectbound(x,y,w,h,7);
		  ps_color(FL_RIGHT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
		  mv(x+w-bw-1.,y+bw+1.); vv(x+w-bw-1.,y+h-bw-1.);
		  vv(x+w-1.,y+h-1.); vv(x+w-1.,y+1.);
		  fprintf(_fl_ps, "closepath\nfill\n");
		  ps_color(FL_BOT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
		  mv(x+1.5,y+1.); vv(x+w-1.,y+1.);
		  vv(x+w-bw-1.,y+bw); vv(x+bw-0.,y+bw);
		  fprintf(_fl_ps, "closepath\nfill\n");
		  ps_rectbound(x+bw,y+bw,w-2*bw,h-2*bw,7);
		  ps_line(x, y+h-1, bw+1, -bw+1, 0);
		  ps_bound(x,y,w,h,0);
		  break;
		case FL_DOWN_BOX:
		  ps_rectbound(x,y,w,h,7);
		  ps_color(FL_RIGHT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
		  mv(x+bw,y+bw+1.); vv(x+bw,y+h-bw);
		  vv(x+1.,y+h-1.); vv(x+1.,y+1.);
		  fprintf(_fl_ps, "closepath\nfill\n");
		  ps_color(FL_BOT_BOUND_COL);
		  fprintf(_fl_ps, "newpath\n");
		  mv(x+1.5,y+h-1.); vv(x+w-1.,y+h-1.);
		  vv(x+w-bw-1.,y+h-bw); vv(x+bw-0.,y+h-bw);
		  fprintf(_fl_ps, "closepath\nfill\n");
		  ps_rectbound(x+bw,y+bw,w-2*bw,h-2*bw,7);
		  ps_line(x+w-bw-1, y+bw, bw+1, -bw+1, 0);
		  ps_bound(x,y,w,h,0);
		  break;
		case FL_NO_BOX:
		  break;
		default:
		  fl_error("fl_ps_box","Unknown box type");
		  break;
	  }
	  break;
  }
}

/*********
   Drawing text.
*********/

#define NL	10

/***
  Major text drawing routine
 ***/

void fl_ps_string(
  int horalign,		/* Horizontal alignment -1=left, 0=center, 1=right */
  int vertalign,	/* Vertical alignment -1=bottom, 0=center, 1=top */
  float x, float y, float w, float h,	/* Bounding box */
  int clip,		/* Whether to clip to bounding box */
  int backcol,		/* Background color */
  int forecol,		/* Foreground color (text color) */
  int curscol,		/* Cursor color */
  float size,		/* Font size */
  int style,		/* Font style */
  int curspos,		/* The cursor position (-1 = no cursor) */
  int selstart,int selend,  /* The selected region (end<start = no region) */
  char str[])		/* The (muli-line) string */
{
  int i;		/* counter */
  char tc;		/* temporary character */
  float tt;		/* temp value */
  char *lines[1024];	/* pointers to the start of the different lines */
  int start[1024];	/* start position of these lines */
  float startx[1024];	/* start x-coordinate of these lines */
  float starty[1024];	/* start y-coordinate of these lines */
  float width[1024];	/* string width of the lines */
  int lnumb;		/* number of lines */
  fmfonthandle fnt;	/* font to be used */
  fmfontinfo info;	/* info of the font */
  float height;		/* height of the font */
  float desc;		/* descender of the font */
  int slstart,slend;	/* Temporary selection positions */
  float xsel,wsel;	/* position and width of selection area */

  /* This routines relies on fm font manager for all string positioning.
	 Postscript can actually calculate this at run time, too. Would be more
	 accurate, probably, but for now this is a decent effort. */

  if (curspos != 0 && (str == NULL || str[0] == '\0')) return;

  /* Set the font to be used and get some info */
  fnt = use_font(style,size);
  ps_use_font(style, size);
  if (fnt == NULL) return;
  fmsetfont(fnt);
  fmgetfontinfo(fnt,&info);
  height = info.ysize;
  desc = info.yorig;

  /* Set clipping if required */
  if (clip) ps_set_clipping(x,y,w,h);

  /* Split string into lines */
  lines[0] = &(str[0]);
  start[0] = 0;
  lnumb = 1;
  for (i = 0; str[i] != '\0'; i++)
    if (str[i] == NL)
    {
      str[i] = '\0';
      lines[lnumb] = &(str[i+1]);
      start[lnumb] = i+1;
      lnumb++;
    }
  start[lnumb] = i+1;

  /* Calculate start coordinates of lines */
  for (i = 0; i < lnumb; i++)
  {
    width[i] = fmgetstrwidth(fnt,lines[i]);
    if (horalign == -1) startx[i] = x;
    if (horalign == 0)  startx[i] = x + 0.5*(w-width[i]);
    if (horalign == 1)  startx[i] = x + w - width[i];
    if (vertalign == -1)  starty[i] = y - (i - lnumb)*height;
    if (vertalign == 0)  starty[i] = y + 0.5*h - (i - 0.5*lnumb)*height;
    if (vertalign == 1) starty[i] = y + h - i*height;
  }

  /* Draw the lines */
  for (i = 0; i < lnumb; i++)
  {
    /* Check whether visible */
    if (clip && starty[i]-height > y+h) continue;
    if (clip && starty[i] < y) break;
    /* Draw it */
	fprintf(_fl_ps, "%f %f moveto\n", startx[i],starty[i]-height+desc);
    ps_color(forecol);
	fprintf(_fl_ps, "(%s) show\n", lines[i]);
    /* Draw selection area if required. */
    if (selstart < start[i+1] && selend >start[i])
    {
      if (selstart <= start[i]) slstart = start[i]; else slstart = selstart;
      if (selend >= start[i+1]) slend = start[i+1]-1; else slend = selend;
      tc = str[slstart];
      str[slstart] = '\0';
      xsel = startx[i] + fmgetstrwidth(fnt,lines[i]);
      str[slstart] = tc;
      tc = str[slend];
      str[slend] = '\0';
      wsel = fmgetstrwidth(fnt,&(str[slstart]));
      ps_rect(xsel,starty[i]-height,wsel,height,forecol);
	  fprintf(_fl_ps, "%f %f moveto\n", xsel,starty[i]-height+desc);
      ps_color(backcol);
	  fprintf(_fl_ps, "(%s) show\n", &(str[slstart]));
      str[slend] = tc;
    }
  }

  /* Draw the cursor */
  if (curspos >= 0)
  {
    i=0; while (start[i] <= curspos) i++;
    i--;
    tc = str[curspos];
    str[curspos] = '\0';
    tt = fmgetstrwidth(fnt,lines[i]);
    str[curspos] = tc;
    ps_rect(startx[i]+tt,starty[i]-height,2.,height,curscol);
  }
    
  /* Restore the original string. */
  for (i = 1; i < lnumb; i++)
    str[start[i]-1] = NL;

  /* Reset clipping if required */
  if (clip) ps_unset_clipping();
}

/***
  Misselaneous text drawing routines
***/

void fl_ps_text_cursor(int renderstyle,
			int align,float x,float y,float w,float h,
			int c,float size,int style,char str[],int cc,int pos)
/* Renders a (multi-line) text with a cursor. */
{
  int horalign,vertalign;
  float height = fl_get_char_height(size,style);

  switch (align) {
     case FL_ALIGN_LEFT:	horalign = -1; vertalign =  0; break;
     case FL_ALIGN_RIGHT:	horalign =  1; vertalign =  0; break;
     case FL_ALIGN_CENTER:	horalign =  0; vertalign =  0; break;
     case FL_ALIGN_TOP:		horalign =  0; vertalign =  1; break;
     case FL_ALIGN_BOTTOM:	horalign =  0; vertalign = -1; break;
  }
  if (renderstyle == FL_POSTSCRIPT)
	  fl_ps_string(horalign,vertalign,
			x+0.5*height,y+0.25*height,w-height,h-0.5*height,
			FALSE, 7,c,cc, size,style, pos, 0,-1, str);
  else
	  fl_ps_string(horalign,vertalign,
			x+0.5*height,y+0.25*height,w-height,h-0.5*height,
			FALSE, 7,0,0, size,style, pos, 0,-1, str);
}

void fl_ps_text(int renderstyle, int align,float x,float y,float w,float h,
	      int c,float size,int style,char str[])
/* Renders a text inside a box. */
{
  if (str == NULL || str[0] == '\0') return;
  if (str[0] == '@') if (fl_ps_symbol(renderstyle, str,x,y,w,h,c)) return;
  fl_ps_text_cursor(renderstyle, align,x,y,w,h,c,size,style,str,0,-1);
}

void fl_ps_text_beside(int renderstyle,
			int align,float x,float y,float w,float h,
			int c,float size,int style, char str[])
/* Renders a text besides a box. */
{
  if (str == NULL || str[0] == '\0') return;
  if      (align == FL_ALIGN_LEFT)
      fl_ps_text(renderstyle, FL_ALIGN_RIGHT,x-h,y,h,h,c,size,style,str);
  else if (align == FL_ALIGN_RIGHT)
      fl_ps_text(renderstyle, FL_ALIGN_LEFT,x+w,y,h,h,c,size,style,str);
  else if (align == FL_ALIGN_TOP)
      fl_ps_text(renderstyle, FL_ALIGN_BOTTOM,x,y+h,w,h,c,size,style,str);
  else if (align == FL_ALIGN_BOTTOM)
      fl_ps_text(renderstyle, FL_ALIGN_TOP,x,y-h,w,h,c,size,style,str);
  else
      fl_ps_text(renderstyle, FL_ALIGN_CENTER,x,y,w,h,c,size,style,str);
}

int fl_ps_symbol(int renderstyle,
	char label[],float x,float y,float w,float h, int col)
/* Renders the symbol with the given label, returns whether it exists */
{
  int pos , i, shift;
  char str[50];
  int rotangle = 0;
  int equalscale = 0;
  if (label[0] != '@') return FALSE;
  if (label[1] == '#') {equalscale = 1; pos = 2;} else pos = 1;
  shift = pos;
  if (label[pos] >= '1' && label[pos] <= '9') {
    switch (label[pos]) {
    case '9': rotangle =  45; break;
    case '8': rotangle =  90; break;
    case '7': rotangle = 135; break;
    case '4': rotangle = 180; break;
    case '1': rotangle = 225; break;
    case '2': rotangle = 270; break;
    case '3': rotangle = 315; break;
    }
    shift = pos+1;
  } else if (label[pos] == '0') {
    rotangle = 100*(label[pos+1]-'0') + 10*(label[pos+2]-'0') +
		    1*(label[pos+3]-'0');
    shift = pos+4;
  }
  i = 0;
  do str[i] = label[i+shift]; while (str[i++] != NULL);
  pos = find(str);
  if (pos == -1) return FALSE;

	fprintf(_fl_ps, "gsave\n");
	fprintf(_fl_ps, "%f %f translate\n", x+w/2.0,y+h/2.0);
	if (symbols[pos].scalable) {
		if (equalscale) {
			if (w<h) {
				fprintf(_fl_ps, "%f %f scale\n", 0.4*w,0.4*w);
				fprintf(_fl_ps, "%f setlinewidth\n", 1.0/(0.4*w),1.0/(0.4*w));
			} else {
				fprintf(_fl_ps, "%f %f scale\n", 0.4*h, 0.4*h);
				fprintf(_fl_ps, "%f setlinewidth\n", 1.0/(0.4*h),1.0/(0.4*h));
			}
		} else {
			fprintf(_fl_ps, "%f %f scale\n", 0.4*w, 0.4*h);
			fprintf(_fl_ps, "%f setlinewidth\n", 1.0/(0.4*w),1.0/(0.4*h));
		}
		fprintf(_fl_ps, "%d rotate\n", rotangle);
	}
	(symbols[pos].renderit)(renderstyle, col);
	fprintf(_fl_ps, "grestore\n");

  return TRUE;
}
